#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MAXCOLOR 23

Display *display;
int screen;
Window window, win2;
Colormap cmap;
XSizeHints hints, hints2;
GC gc_color[MAXCOLOR+1];
XColor color[MAXCOLOR+1];
unsigned long black_pixel, white_pixel;
unsigned int depth;

char *vari[MAXCOLOR] = {
  "black",
  "white",
  "yellow",
  "green",
  "red",
  "blue",
  "gray10",
  "gray20",
  "gray30",
  "gray40",
  "gray50",
  "gray60",
  "gray70",
  "gray80",
  "brown",
  "gray10",
  "gray20",
  "gray30",
  "gray40",
  "gray50",
  "gray60",
  "gray70",
  "gray80"
};

main(int argc, char **argv) {
  char *display_name=NULL;
  int x=1, y=1;
  
  init_x (display_name, argc, argv);
  do_frac ();
  XSync (display, 1);
  end_x();
}

/*****************************************************************************
*                                                                            *
* The following code is purely X-Windows version 11 release 3-5 dependent    *
*                                                                            *
*****************************************************************************/

int init_x (char *disp_name, int argc, char *argv)
{
    Window root_window;
    XColor exact;
    XEvent ev;
    int i;
    char varinimi[80];

    display = XOpenDisplay (disp_name);
    if (display == (Display *)NULL)
    {
	fprintf (stderr, "Cannot connect to X11-server!\n");
	exit (-1) ;
    }
    screen = DefaultScreen (display);
    root_window = DefaultRootWindow (display);

    depth = DefaultDepth (display, screen);
    cmap = DefaultColormap (display, screen);
    black_pixel = BlackPixel (display, screen);
    white_pixel = WhitePixel (display, screen);
    for (i = 0; i < MAXCOLOR; i++) {
/*      sprintf (varinimi, "gray%d", (int)((float)i*(100.0/(float)MAXCOLOR))); */
      strcpy (varinimi, vari[i]);
      if (!XAllocNamedColor (display, cmap, varinimi, &exact, &color[i])) {
	fprintf (stderr, "Failed to allocate colors!\n");
	exit (1);
      }
    }

    hints.x = hints.y = 50;
    hints2.x = hints.x + 400;
    hints2.y = hints.y;
    hints2.width = hints.width = 400;
    hints2.height = hints.height = 400;
    hints2.flags = hints.flags = PPosition | PSize ;
    window = XCreateSimpleWindow (display, root_window, hints.x, hints.y,
				  hints.width, hints.height, 2,
				  color[1].pixel,  black_pixel);
    XSetStandardProperties (display, window, "xfrac by tommi@phoenix.oulu.fi", "simplex",
			    None, (char**)argv, argc, &hints);
    win2 = XCreateSimpleWindow (display, root_window, hints2.x, hints2.y,
				  hints2.width, hints2.height, 2,
				  color[1].pixel,  black_pixel);
    XSetStandardProperties (display, win2, "xfrac by tommi@phoenix.oulu.fi", "simplex",
			    None, (char**)argv, argc, &hints2);

    for (i=0; i< MAXCOLOR; i++) {
      gc_color[i] = XCreateGC (display, window, 0L, 0);
      XSetForeground (display, gc_color[i], color[i].pixel);
      XSetBackground (display, gc_color[i], black_pixel);
    }

    XSelectInput (display, window, ExposureMask);
    XMapRaised (display, window);
    XMapRaised (display, win2);

    do XNextEvent (display, &ev);
    while (ev.type != Expose);

    XSelectInput (display, window, ExposureMask | ButtonPressMask |
                  ButtonReleaseMask | KeyPressMask);
    return 1;
}


int end_x()
{
    XSync (display, 1);
    XCloseDisplay (display);
    return 1;
}

int do_frac() {
  double rr, ri, zr, zi, cr, ci, dr, di, mag;
  int x, y, n, iter;
  XEvent event;
  int numEvents, pointerX, pointerY, scaleX0, scaleY0, scaleX1, scaleY1;
  int showiter, proceed, update_mandelbrot;

  dr = di = 0;
  mag = 1;
  iter = 100;
  pointerX = pointerY = 200;
  update_mandelbrot = 1;
  for (;;) {
    if (update_mandelbrot) {
      for (y = -200; y <= 200; y++) {
        for (x = -200; x <= 200; x++) {
          rr = ri = zi = zr = 0.0;
          cr = ((double)x / 100.0)*mag + dr;
          ci = ((double)y / 100.0)*mag + di;
          for (n = 0; n < iter; n++) {

            rr = zr*zr - zi*zi + cr;
            zi = 2*zr*zi + ci;
            zr = rr;
            if (zr*zr + zi*zi > 4.0) break;
          }
          XDrawPoint (display, window, gc_color[(iter-n) % MAXCOLOR], 200+x, 200+y);
        }
      }
    }
    XSync (display, False);
    update_mandelbrot = showiter = proceed = 0;
    do {
      numEvents = XEventsQueued (display, QueuedAfterReading);
      while (numEvents) {
        XNextEvent (display, &event);
        switch(event.type) {
          case ButtonPress:
            switch (event.xbutton.button) {
              case Button1:
                pointerX = (int)event.xbutton.x;
                pointerY = (int)event.xbutton.y;
                proceed = showiter = 1;
                break;
              case Button2:
                scaleX0 = (int)event.xbutton.x;
                scaleY0 = (int)event.xbutton.y;
                break;
              case Button3:
                mag = 1;
                fprintf (stderr, "dr=%le, di=%le, mag=%le\n", dr, di, mag);
                update_mandelbrot = proceed = 1;
                break;
            }
            break;
          case ButtonRelease:
            if (event.xbutton.button == Button2) {
                scaleX1 = (int)event.xbutton.x;
                scaleY1 = (int)event.xbutton.y;
                dr = dr + (double)(scaleX1+scaleX0-400)/200.0*mag;
                di = di + (double)(scaleY1+scaleY0-400)/200.0*mag;
                mag = (mag/400.0)*((double)scaleX1-(double)scaleX0);
                fprintf (stderr, "dr=%le, di=%le, mag=%le\n", dr, di, mag);
                update_mandelbrot = proceed = 1;
            }
            break;
          case MotionNotify:
            break;
          case KeyPress:
            break;
        }
        numEvents--;
      }
    } while (!proceed);
    
    if (showiter) {
      cr = ((double)(pointerX-200) / 100.0)*mag + dr;
      ci = ((double)(pointerY-200) / 100.0)*mag + di;
      XFillRectangle (display, win2, gc_color[0], 0, 0, 400, 400);
      for (y = -200; y <= 200; y++) {
        for (x = -200; x <= 200; x++) {
          rr = ri = 0;
          zr = ((double)x / 100.0)*mag + dr;
          zi = ((double)y / 100.0)*mag + di;
          for (n = 0; n < iter; n++) {
            rr = zr*zr - zi*zi + cr;
            zi = 2*zr*zi + ci;
            zr = rr;
            if (zr*zr + zi*zi > 4.0) break;
          }
          XDrawPoint (display, win2, gc_color[(iter-n) % MAXCOLOR], 200+x, 200+y);
        }
      }
      XSync (display, False);
    }
  }
}
